from __future__ import print_function

import sys, time
import datetime


# Functions to detect whether we are running inside a notebook or not

from IPython import get_ipython


def is_inside_notebook():

    ip = get_ipython()

    if ip is None:

        # This happens if we are running in a python session, not a IPython one (for example in a script)
        return False

    else:

        # We are running in a IPython session, either in a console or in a notebook
        if ip.has_trait("kernel"):

            # We are in a notebook
            return True

        else:

            # We are not in a notebook
            return False


# This is used for testing purposes

test_ascii_only = False

try:

    from ipywidgets import FloatProgress, HTML, VBox

except ImportError:

    has_widgets = False

else:

    if test_ascii_only:

        has_widgets = False

    else:

        has_widgets = True


def fallback_display(x):

    print(x)


try:

    from IPython.display import display

except ImportError:

    display = fallback_display


from contextlib import contextmanager


class CannotGenerateHTMLBar(RuntimeError):
    pass


@contextmanager
def progress_bar(iterations, width=None, scale=1, units="", title=None):
    """
    Use as a context manager to display a progress bar which adapts itself to the environment. It will be a widget
    in jupyter or a text progress bar in a terminal (but you don't have to worry about it)
    :param iterations: number of iterations for completion of the task
    :param width: width of the progress bar (default: None, which means self-decided)
    :param scale: display the progress scaled for this number (useful to display downloads for example) (default: 1)
    :param units: a unit to display after the progress (default: '')
    :param title: a title for the task, which will be displayed before the progress bar (default: None, i.e., no title)
    :return: a ProgressBarAscii or a ProgressBarHTML instance, depending on the environment (jupyter or terminal)
    """

    # Instance progress bar

    if has_widgets and is_inside_notebook():

        try:

            if width is None:

                bar_width = 50

            else:

                bar_width = int(width)

                # Default is the HTML bar, which only works within a notebook

            this_progress_bar = ProgressBarHTML(
                iterations, bar_width, scale=scale, units=units, title=title
            )

        except:

            # Fall back to Ascii progress bar

            if width is None:

                bar_width = 30

            else:

                bar_width = int(width)

            # Running in a terminal. Fall back to the ascii bar

            this_progress_bar = ProgressBarAscii(
                iterations, bar_width, scale=scale, units=units, title=title
            )

    else:

        if width is None:

            bar_width = 30

        else:

            bar_width = int(width)

        # No widgets available, fall back to ascii bar

        this_progress_bar = ProgressBarAscii(
            iterations, bar_width, scale=scale, units=units, title=title
        )

    yield this_progress_bar  # type: ProgressBarBase

    this_progress_bar.finish()


class ProgressBarBase(object):
    def __init__(self, iterations, width, scale=1, units="", title=None):

        # Store the number of iterations

        self._iterations = int(iterations)

        # Store the width (in characters)

        self._width = width

        # Get the start time

        self._start_time = time.time()

        # Current iteration is zero
        self._last_iteration = 0

        # last printed percent
        self._last_printed_percent = 0

        # Store the scale
        self._scale = float(scale)

        # store the units
        self._units = units

        # Store the title
        self._title = title

        # Setup

        self._setup()

    def _setup(self):

        raise NotImplementedError("Need to override this")

    def animate(self, iteration):

        # We only update the progress bar if the progress has gone backward,
        # or if the progress has increased by at least 1%. This is to avoid
        # updating it too much, which would fill log files in text mode,
        # or slow down the computation in HTML mode

        this_percent = iteration / float(self._iterations) * 100.0

        if (
            this_percent - self._last_printed_percent < 0
            or (this_percent - self._last_printed_percent) >= 1
        ):

            self._last_iteration = self._animate(iteration)

            self._last_printed_percent = this_percent

        else:

            self._last_iteration = iteration

    def _animate(self, iteration):

        raise NotImplementedError("Need to override this")

    def increase(self, n_steps=1):

        self.animate(self._last_iteration + n_steps)

    def finish(self):

        self._animate(self._iterations)

    def _check_remaining_time(self, current_iteration, delta_t):

        if current_iteration == 0:
            return "--:--"

        # Seconds per iterations
        s_per_iter = delta_t / float(current_iteration)

        # Seconds to go (estimate)
        s_to_go = s_per_iter * (self._iterations - current_iteration)

        # I cast to int so it won't show decimal seconds

        return str(datetime.timedelta(seconds=int(s_to_go)))

    def _get_label(self, current_iteration):

        delta_t = time.time() - self._start_time

        elapsed_iter = min(current_iteration, self._iterations)

        if self._scale != 1:

            label_text = "%.2f / %.2f %s in %.1f s (%s remaining)" % (
                elapsed_iter / self._scale,
                self._iterations / self._scale,
                self._units,
                delta_t,
                self._check_remaining_time(current_iteration, delta_t),
            )

        else:

            label_text = "%d / %s %s in %.1f s (%s remaining)" % (
                elapsed_iter,
                self._iterations,
                self._units,
                delta_t,
                self._check_remaining_time(current_iteration, delta_t),
            )

        return label_text


class ProgressBarHTML(ProgressBarBase):
    def __init__(self, iterations, width, scale=1, units="", title=None):
        super(ProgressBarHTML, self).__init__(
            iterations, width, scale, units=units, title=title
        )

    def _setup(self):
        # Setup the widget, which is a bar between 0 and 100

        self._bar = FloatProgress(min=0, max=100)

        # Set explicitly the bar to 0

        self._bar.value = 0

        # Setup also an HTML label (which will contain the progress, the elapsed time and the foreseen
        # completion time)

        self._title_cell = HTML()

        if self._title is not None:
            self._title_cell.value = "%s : " % self._title

        self._label = HTML()
        self._vbox = VBox(children=[self._title_cell, self._label, self._bar])

        # Display everything

        display(self._vbox)

        self._animate(0)

    def _animate(self, iteration):
        current_label = self._get_label(iteration)

        self._bar.value = float(iteration) / float(self._iterations) * 100

        self._label.value = current_label

        return iteration


class ProgressBarAscii(ProgressBarBase):
    def __init__(self, iterations, width, scale=1, units="", title=None):
        super(ProgressBarAscii, self).__init__(
            iterations, width, scale, units, title=title
        )

    def _setup(self):
        self._fill_char = "*"

        # Display the title
        print("%s :\n" % self._title)

        # Display an empty bar
        self._animate(0)

    def _animate(self, current_iteration):
        current_bar = self._generate_bar(current_iteration)
        current_label = self._get_label(current_iteration)

        print("\r%s  %s" % (current_bar, current_label), end="")
        sys.stdout.flush()

        return current_iteration

    def finish(self):
        super(ProgressBarAscii, self).finish()

        sys.stdout.write("\n")

    def _generate_bar(self, current_iteration):
        # Compute the percentage completed

        elapsed_iter = min(current_iteration, self._iterations)

        new_amount = (elapsed_iter / float(self._iterations)) * 100.0

        percent_done = min(int(round((new_amount / 100.0) * 100.0)), 100)

        # Generate the bar

        all_full = self._width - 2

        num_hashes = int(round((percent_done / 100.0) * all_full))

        bar = "[" + self._fill_char * num_hashes + " " * (all_full - num_hashes) + "]"

        # Now place the completed percentage in the middle of the bar

        pct_place = (len(bar) // 2) - len(str(percent_done))
        pct_string = "%d%%" % percent_done

        bar = bar[0:pct_place] + (pct_string + bar[pct_place + len(pct_string) :])

        return bar



import numpy as np
import matplotlib.pyplot as plt

from pychangcooper.utils.cmap_intervals import cmap_intervals


def fill_plot_static(x, y_values, cmap="viridis", alpha=1.0, ax=None):
    """
    
    plot all the y values 
    
    :param x: the x values 
    :param y_values: the matrix of y values
    :param cmap: mpl cmap
    :param alpha: the transparency 
    :param ax: optional ax
    :return: mpl figure
    """

    # create an axis and fig if not provided
    if ax is None:
        fig, ax = plt.subplots()

    else:

        fig = ax.get_figure()

    # create a list of colors

    colors = cmap_intervals(len(y_values), cmap)

    # keep figures up front

    zorder = len(y_values) + 10

    # loop through the y matrix and plot

    for i, y in enumerate(y_values):

        plot_time_step(i, x, y_values, colors[i], zorder, alpha, ax)

        # decrease the z order

        zorder -= 1

    return fig


def plot_time_step(iteration, x, y_values, color, zorder, alpha, ax):
    """
    plot an individual time step with a filled plot
    
    :param iteration: the iteration to consider
    :param x: the x values
    :param y_values: the matrix of y vaules
    :param color: the color to plot
    :param zorder: the zorder of the plot
    :param alpha: the transparency
    :param ax: the ax to plot with
    :return: None
    """

    # the first iteration is a simple line

    if iteration == 0:

        ax.plot(x, y_values[iteration], alpha=alpha, zorder=zorder, color=color)

    # don't fill to zero if there is nothing to plot

    elif np.all(y_values[iteration - 1] == 0.0):

        ax.plot(x, y_values[iteration], alpha=alpha, zorder=zorder, color=color)

    # fill between this iteration and the last
    else:

        ax.fill_between(
            x,
            y_values[iteration - 1],
            y_values[iteration],
            alpha=alpha,
            zorder=zorder,
            color=color,
        )



import numpy as np

from pychangcooper.chang_cooper import ChangCooper, log_grid_generator
from pychangcooper.photons.photon_emitter import PhotonEmitter
from pychangcooper.photons.synchrotron_emission import (
    SynchrotronEmission,
    synchrotron_cooling_constant,
    synchrotron_cooling_time,
)
from pychangcooper.scenarios.continuous_powerlaw_injection import (
    ContinuousPowerlawInjection,
)
from pychangcooper.scenarios.generic_cooling_component import GenericCoolingComponent


class SynchrotronCoolingComponent(GenericCoolingComponent):
    def __init__(self, B):
        """
        A synchrotron cooling component that implements the proper cooling terms
        
        :param B: magnetic field strength in Gauss
        
        """

        C0 = synchrotron_cooling_constant(B)

        super(SynchrotronCoolingComponent, self).__init__(C0, 2.0)


class SynchrotronCooling_ImpulsivePLInjection(
    SynchrotronCoolingComponent, PhotonEmitter, ChangCooper
):
    def __init__(
        self,
        B=10.0,
        index=-2.2,
        gamma_injection=1e3,
        gamma_cool=2e3,
        gamma_max=1e5,
        n_grid_points=300,
        max_grid=1e7,
        store_progress=False,
    ):
        """
        Synchrotron cooling and radiation from an impulsive power law injection of particles.
        
        :param B: the magnetic field strength in Gaussa
        :param index: the injected electron spectral index
        :param gamma_injection: the electron injection energy
        :param gamma_cool: the cooling energy
        :param gamma_max: the maximum injected energy
        :param n_grid_points: the number of grid point
        :param max_grid: the maximum grid energy
        :param store_progress: to store progress
        """

        self._B = B

        bulk_gamma = 300.0

        # calculate the number of steps to cool the maximum
        # electron energy to gamma cool

        ratio = gamma_max / gamma_cool

        n_steps = np.round(ratio)

        # assign the properties to the calss

        self._gamma_max = gamma_max
        self._gamma_cool = gamma_cool
        self._gamma_injection = gamma_injection
        self._index = index

        # build the initial power law injection

        initial_distribution = np.zeros(n_grid_points)
        tmp_grid, _, _ = log_grid_generator(n_grid_points, max_grid)

        idx = (gamma_injection <= tmp_grid) & (tmp_grid <= gamma_max)

        # normalize the electrons so that N = number of electrons

        norm = (
            np.power(gamma_max, index + 1) - np.power(gamma_injection, index + 1)
        ) / (index + 1)

        initial_distribution[idx] = 1.0 * norm * np.power(tmp_grid[idx], index)

        # set the time step to be that of the cooling time of
        # the most energetic electrons

        delta_t = synchrotron_cooling_time(B, gamma_max)

        # initialize the cooling terms

        SynchrotronCoolingComponent.__init__(self, B)

        # build the solver

        ChangCooper.__init__(
            self, n_grid_points, max_grid, delta_t, initial_distribution, store_progress
        )

        # create the emission kernel for producing radiation

        emission_kernel = SynchrotronEmission(self._grid, self._B)

        # initialize the radiation production

        PhotonEmitter.__init__(self, n_steps, emission_kernel)


class SynchrotronCooling_ContinuousPLInjection(
    SynchrotronCoolingComponent, ContinuousPowerlawInjection, PhotonEmitter, ChangCooper
):
    def __init__(
        self,
        B=10.0,
        index=-2.2,
        gamma_injection=1e3,
        gamma_cool=2e3,
        gamma_max=1e5,
        n_grid_points=300,
        max_grid=1e7,
        store_progress=False,
    ):
        """
        Synchrotron cooling and radiation from continuous power law injection of particles.
        
        :param B: the magnetic field strength in Gaussa
        :param index: the injected electron spectral index
        :param gamma_injection: the electron injection energy
        :param gamma_cool: the cooling energy
        :param gamma_max: the maximum injected energy
        :param n_grid_points: the number of grid point
        :param max_grid: the maximum grid energy
        :param store_progress: to store progress
        """

        self._B = B

        bulk_gamma = 300.0

        # calculate the number of steps to cool the maximum
        # electron energy to gamma cool

        ratio = gamma_max / gamma_cool

        n_steps = np.round(ratio)

        # assign the properties to the calss

        self._gamma_max = gamma_max
        self._gamma_cool = gamma_cool
        self._gamma_injection = gamma_injection
        self._index = index

        # set the time step to be that of the cooling time of
        # the most energetic electrons

        delta_t = synchrotron_cooling_time(B, gamma_max)
        # initialize the cooling terms

        SynchrotronCoolingComponent.__init__(self, B)

        # create the injection
        ContinuousPowerlawInjection.__init__(
            self, gamma_injection, gamma_max, index, N=1.0
        )

        # build the solver

        ChangCooper.__init__(
            self, n_grid_points, max_grid, delta_t, None, store_progress
        )

        # build the synchrotron emission kernel

        emission_kernel = SynchrotronEmission(self._grid, self._B)

        # initialize the photon emission process

        PhotonEmitter.__init__(self, n_steps, emission_kernel)

    def _clean(self):

        if self._iterations <= 1:

            self._idx_max = np.argmax(self._n_current)
            self._sync_max = self._n_current[self._idx_max]

        # now clean anything below

        idx1 = self._grid < self._gamma_injection

        idx2 = self._n_current < self._sync_max

        idx3 = idx1 & idx2

        self._n_current[idx3] = 0.0

        # now handle cooling

        if self._gamma_cool < self._gamma_injection:

            idx = self._grid < self._gamma_cool

            self._n_current[idx] = 0.0

        else:

            idx = self._grid < self._gamma_injection

            self._n_current[idx] = 0.0

        # lower_bound = min(self._gamma_cool, self._gamma_injection)

        # lower_bound = self._gamma_cool


#        idx = self._grid <= lower_bound

# self._n_current[:idx] = 0.



import matplotlib.pyplot as plt
import numpy as np


# reverse these colormaps so that it goes from light to dark

REVERSE_CMAP = ["summer", "autumn", "winter", "spring", "copper"]

# clip some colormaps so the colors aren't too light

CMAP_RANGE = dict(
    gray={"start": 200, "stop": 0},
    Blues={"start": 60, "stop": 255},
    Oranges={"start": 100, "stop": 255},
    OrRd={"start": 60, "stop": 255},
    BuGn={"start": 60, "stop": 255},
    PuRd={"start": 60, "stop": 255},
    YlGn={"start": 60, "stop": 255},
    YlGnBu={"start": 60, "stop": 255},
    YlOrBr={"start": 60, "stop": 255},
    YlOrRd={"start": 60, "stop": 255},
    hot={"start": 230, "stop": 0},
    bone={"start": 200, "stop": 0},
    pink={"start": 160, "stop": 0},
)


_alt_cmaps = {
    "vaporwave": [
        "#94D0FF",
        "#8795E8",
        "#966bff",
        "#AD8CFF",
        "#C774E8",
        "#c774a9",
        "#FF6AD5",
        "#ff6a8b",
        "#ff8b8b",
        "#ffa58b",
        "#ffde8b",
        "#cdde8b",
        "#8bde8b",
        "#20de8b",
    ],
    "cool": ["#FF6AD5", "#C774E8", "#AD8CFF", "#8795E8", "#94D0FF"],
    "crystal_pepsi": ["#FFCCFF", "#F1DAFF", "#E3E8FF", "#CCFFFF"],
    "mallsoft": ["#fbcff3", "#f7c0bb", "#acd0f4", "#8690ff", "#30bfdd", "#7fd4c1"],
    "jazzcup": ["#392682", "#7a3a9a", "#3f86bc", "#28ada8", "#83dde0"],
    "sunset": ["#661246", "#ae1357", "#f9247e", "#d7509f", "#f9897b"],
    "macplus": ["#1b4247", "#09979b", "#75d8d5", "#ffc0cb", "#fe7f9d", "#65323e"],
    "seapunk": ["#532e57", "#a997ab", "#7ec488", "#569874", "#296656"],
}


def cmap_intervals(length=50, cmap="YlOrBr", start=None, stop=None):
    """
    Return evenly spaced intervals of a given colormap `cmap`.
    Colormaps listed in REVERSE_CMAP will be cycled in reverse order.
    Certain colormaps have pre-specified color ranges in CMAP_RANGE. These module
    variables ensure that colors cycle from light to dark and light colors are
    not too close to white.
    :param length: int the number of colors used before cycling back to first color. When
    length is large (> ~10), it is difficult to distinguish between
    successive lines because successive colors are very similar.
    :param cmap: str name of a matplotlib colormap (see matplotlib.pyplot.cm)
    """

    # qualitative color maps
    if cmap in [
        "Accent",
        "Dark2",
        "Paired",
        "Pastel1",
        "Pastel2",
        "Set1",
        "Set2",
        "Set3",
        "Vega10",
        "Vega20",
        "Vega20b",
        "Vega20c",
    ]:

        cm = getattr(plt.cm, cmap)

        base_n_colors = cm.N

        cmap_list = cm(range(base_n_colors))

        if base_n_colors < length:

            factor = int(np.floor_divide(length, base_n_colors))

            cmap_list = np.tile(cmap_list, (factor, 1))

        return cmap_list

    elif cmap in _alt_cmaps.keys():

        this_cmap = _alt_cmaps[cmap]

        if len(this_cmap) >= length:

            return this_cmap

        else:

            raise RuntimeError("CMAP is too short %d<%d" % (len(this_cmap), length))

    else:

        cm = getattr(plt.cm, cmap)

        crange = CMAP_RANGE.get(cmap, dict(start=0, stop=255))
        if cmap in REVERSE_CMAP:
            crange = dict(start=crange["stop"], stop=crange["start"])
        if start is not None:
            crange["start"] = start
        if stop is not None:
            crange["stop"] = stop

        idx = np.linspace(crange["start"], crange["stop"], length).astype(np.int)
        return cm(idx)



import numpy as np
import matplotlib.pyplot as plt

from pychangcooper.utils.progress_bar import progress_bar
from pychangcooper.io.fill_plot import fill_plot_static


class PhotonEmitter(object):
    def __init__(self, n_steps, emission_kernel):

        self._n_steps = n_steps

        self._emission_kernel = emission_kernel

    def run(self, photon_energies=None):

        with progress_bar(int(self._n_steps), title="solving electrons electrons") as p:
            for i in range(int(self._n_steps)):
                self.solve_time_step()

                p.increase()

        if photon_energies is not None:

            self._compute_spectrum(photon_energies)

    def _compute_spectrum(self, photon_energies):
        """

        """

        self._emission_kernel.set_photon_energies(photon_energies)

        self._all_spectra = []

        with progress_bar(int(self._n_steps), title="computing spectrum") as p:
            for electrons in self.history:
                self._all_spectra.append(
                    self._emission_kernel.compute_spectrum(electrons)
                )

                p.increase()

        self._all_spectra = np.array(self._all_spectra)

        self._total_spectrum = self._all_spectra.sum(axis=0)

    @property
    def final_spectrum(self):

        return self._total_spectrum

    @property
    def photon_energies(self):

        return self._emission_kernel.photon_energies

    def plot_final_emission(self, ax=None, x_scaling=1.0, y_scaling=1.0, **kwargs):

        assert self.photon_energies is not None, "There are no photons!"

        if ax is None:

            fig, ax = plt.subplots()

        else:

            fig = ax.get_figure()

        ax.loglog(
            self.photon_energies * x_scaling,
            self.photon_energies ** 2 * self.final_spectrum * y_scaling,
            **kwargs
        )

        ax.set_xlabel("Energy")
        ax.set_ylabel(r"$\nu F_{\nu}$")

        return fig

    def plot_initial_emission(self, ax=None, **kwargs):

        assert self.photon_energies is not None, "There are no photons!"
        assert np.all(self._all_spectra[0] > 0.0), "There are no photons to start!"

        if ax is None:

            fig, ax = plt.subplots()

        else:

            fig = ax.get_figure()

        ax.loglog(
            self.photon_energies,
            self.photon_energies ** 2 * self._all_spectra[0],
            **kwargs
        )

        ax.set_xlabel("Energy")
        ax.set_ylabel(r"$\nu F_{\nu}$")

        return fig

    def plot_emission(self, cmap="viridis", skip=1, alpha=0.5, ax=None):

        assert self.photon_energies is not None, "There are no photons!"
        cumulative_spectrum = (self._all_spectra.cumsum(axis=0))[::skip]

        fig = fill_plot_static(
            self._emission_kernel.photon_energies,
            self._emission_kernel.photon_energies ** 2 * cumulative_spectrum,
            cmap,
            alpha,
            ax,
        )

        if ax is None:
            ax = fig.get_axes()[0]

        ax.set_xscale("log")
        ax.set_yscale("log")

        ax.set_xlabel("Energy")
        ax.set_ylabel(r"$\nu F_{\nu}$")

        return fig

    def plot_photons_and_electrons(self, cmap="viridis", skip=1, alpha=0.5):

        fig, (ax1, ax2) = plt.subplots(1, 2)

        _ = self.plot_evolution(cmap=cmap, skip=skip, ax=ax1, alpha=alpha)
        _ = self.plot_emission(cmap=cmap, skip=skip, ax=ax2, alpha=alpha)
        ax2.yaxis.tick_right()
        ax2.yaxis.set_label_position("right")

        ax1.set_xlim(left=min(self._gamma_cool, self._gamma_injection) * 0.5)

        fig.tight_layout()
        fig.subplots_adjust(hspace=0, wspace=0)

        return fig



import numpy as np

from pychangcooper.photons.emission_kernel import EmissionKernel
from pychangcooper.photons.synchrotron_kernal import synchrotron_kernel


class SynchrotronEmission(EmissionKernel):
    def __init__(self, gamma_grid, B):
        """
        :param photon_energies: the photon energies
        :param gamma_grid: the electron gamma grid
        :param B: the B-field in units of Gauss
        """

        self._B = B
        self._gamma_grid = gamma_grid
        self._n_grid_points = len(gamma_grid)

        super(SynchrotronEmission, self).__init__()

    def set_photon_energies(self, photon_energies):
        """
        Set the photon energies and build the synchrotron kernel
        """

        super(SynchrotronEmission, self).set_photon_energies(photon_energies)

        self._build_synchrotron_kernel()

    def _build_synchrotron_kernel(self):
        """
        pre build the synchrotron kernel for the integration
        """
        self._synchrotron_kernel = np.zeros(
            (self._n_photon_energies, self._n_grid_points)
        )
        Bcritical = 4.14e13  # Gauss

        ec = 1.5 * self._B / Bcritical

        for i, energy in enumerate(self._photon_energies):
            for j, gamma in enumerate(self._gamma_grid):
                arg = energy / (ec * gamma * gamma)

                self._synchrotron_kernel[i, j] = synchrotron_kernel(arg)

    def compute_spectrum(self, electron_distribution):
        """
        
        :param electron_distribution: the electron distribution to convolve
        :return: 
        """

        spectrum = np.zeros_like(self._photon_energies)

        # convolve the synchrotron kernel with the electron
        # distribution
        for i, energy in enumerate(self._photon_energies):
            spectrum[i] = (
                self._synchrotron_kernel[i, 1:]
                * electron_distribution[1:]
                * (self._gamma_grid[1:] - self._gamma_grid[:-1])
            ).sum() / (2.0 * energy)

        return spectrum


def synchrotron_cooling_constant(B):
    """
    Compute the characteristic cooling constant for synchrotron cooling

    Define a factor that is dependent on the magnetic field B.
    The cooling time of an electron at energy gamma is
    DT = 6 * pi * me*c / (sigma_T B^2 gamma)
    DT = (1.29234E-9 B^2 gamma)^-1 seconds
    DT = (cool * gamma)^-1 seconds
    where B is in Gauss.

    """

    bulk_gamma = 300.0

    const_factor = 1.29234e-9

    C0 = const_factor * B ** 2

    return C0


def synchrotron_cooling_time(B, gamma):
    """
    Compute the characteristic cooling time of an electron of energy gamma
    for a given magnetic field strength

    """
    # get the cooling constant
    C0 = synchrotron_cooling_constant(B)

    return 1.0 / (C0 * gamma)



import numpy as np


class GenericCoolingComponent(object):
    def __init__(self, C0, cooling_index):
        """
        A generic cooling component that must be co-inherited with
        ChangCooper to produce a cooling only solution
        
        :param C0: the cooling constant
        :param cooling_index: the cooling index
        """

        self._cooling_index = cooling_index
        self._C0 = C0

    def _define_terms(self):

        self._dispersion_term = np.zeros(self._n_grid_points)
        self._heating_term = self._C0 * np.power(self._half_grid, self._cooling_index)



import numpy as np


class ContinuousPowerlawInjection(object):
    def __init__(self, x_min, x_max, index, N=1):

        self._x_min = x_min
        self._x_max = x_max
        self._N = N
        self._index = index

        self._norm = (np.power(x_max, index + 1) - np.power(x_min, index + 1)) / (
            index + 1
        )

    def _source_function(self, x):
        """
        power law injection
        
        """

        out = np.zeros(self._n_grid_points)

        idx = (self._x_min <= self._grid) & (self._grid <= self._x_max)

        out[idx] = self._N * self._norm * np.power(self._grid[idx], self._index)
        return out



import numpy as np

from pychangcooper.chang_cooper import ChangCooper, log_grid_generator
from pychangcooper.scenarios.generic_cooling_acceleration import (
    GenericCoolingAccelerationComponent,
)
from pychangcooper.scenarios.continuous_powerlaw_injection import (
    ContinuousPowerlawInjection,
)
from pychangcooper.photons.photon_emitter import PhotonEmitter
from pychangcooper.scenarios.continuous_powerlaw_injection import (
    ContinuousPowerlawInjection,
)
from pychangcooper.photons.synchrotron_emission import (
    SynchrotronEmission,
    synchrotron_cooling_constant,
    synchrotron_cooling_time,
)


class SynchrotronCoolingAccelerationComponent(GenericCoolingAccelerationComponent):
    def __init__(self, B, t_acc, acceleration_index):
        """
        A synchrotron coooling and acceleration component that must be co-inherited with ChangCooper
        to build a full solver
        
        :param B: the magnetic field strength in Gauss
        :param t_acc: the acceleration time scale
        :param acceleration_index: the acceleration index
        """

        # get the cooling constant
        C0 = synchrotron_cooling_constant(B)

        super(SynchrotronCoolingAccelerationComponent, self).__init__(
            C0, t_acc, cooling_index=2, acceleration_index=acceleration_index
        )


class SynchCoolAccel_ImpulsivePLInjection(
    SynchrotronCoolingAccelerationComponent, PhotonEmitter, ChangCooper
):
    def __init__(
        self,
        B=10.0,
        index=-2.2,
        gamma_injection=1e3,
        gamma_cool=2e3,
        gamma_max=1e5,
        t_acc_fraction=1.0,
        acceleration_index=2,
        n_grid_points=300,
        max_grid=1e7,
        store_progress=False,
    ):

        """
            A synchrotron cooling and acceleration solver and radiation producer with and impulsive
            injection of power law distributed electrons. The acceleration timescale is taken as a 
            fraction of the cooling timescale. The acceleration index expresses the turbulence distribution
            of the magnetic field: 1 for Bohm, 5/3 for Kolmogorov, 2 for Fermi.
            
            
            :param B: the magnetic field strength in Gaussa
            :param index: the injected electron spectral index
            :param gamma_injection: the electron injection energy
            :param gamma_cool: the cooling energy
            :param gamma_max: the maximum injected energy
            :param t_acc_fraction: the acceleration time relative to the cooling time
            :param acceleration_index: the acceleration index
            :param n_grid_points: number of grid points
            :param max_grid: the maximum grid energy
            :param store_progress: to store the progress
            """

        self._B = B

        bulk_gamma = 300.0

        # calculate the number of steps to cool the maximum
        # electron energy to gamma cool

        ratio = gamma_max / gamma_cool

        n_steps = np.round(ratio)

        # assign the properties to the calss

        self._gamma_max = gamma_max
        self._gamma_cool = gamma_cool
        self._gamma_injection = gamma_injection
        self._index = index

        # build the initial power law injection

        initial_distribution = np.zeros(n_grid_points)
        tmp_grid, _, _ = log_grid_generator(n_grid_points, max_grid)

        idx = (gamma_injection <= tmp_grid) & (tmp_grid <= gamma_max)

        # normalize the electrons so that N = number of electrons

        norm = (
            np.power(gamma_max, index + 1) - np.power(gamma_injection, index + 1)
        ) / (index + 1)

        initial_distribution[idx] = 1.0 * norm * np.power(tmp_grid[idx], index)

        # calculate the cooling time of the highest energy electron
        # and set the acceleration time to the fraction of that

        cooling_time = synchrotron_cooling_time(B, gamma_max)

        t_acc = cooling_time * t_acc_fraction

        # initialize the cooling and acceleration terms

        SynchrotronCoolingAccelerationComponent.__init__(
            self, B, t_acc, acceleration_index
        )

        # make the time step the small enough to account for either the cooling
        # or the acceleration

        delta_t = min(t_acc, cooling_time)

        # initialize the solver

        ChangCooper.__init__(
            self, n_grid_points, max_grid, delta_t, initial_distribution, store_progress
        )

        # create the synchrotron emission kernel

        emission_kernel = SynchrotronEmission(self._grid, self._B)

        # initialize the photon emission constructor

        PhotonEmitter.__init__(self, n_steps, emission_kernel)


class SynchCoolAccel_ContinuousPLInjection(
    SynchrotronCoolingAccelerationComponent,
    ContinuousPowerlawInjection,
    PhotonEmitter,
    ChangCooper,
):
    def __init__(
        self,
        B=10.0,
        index=-2.2,
        gamma_injection=1e3,
        gamma_cool=2e3,
        gamma_max=1e5,
        t_acc_fraction=1.0,
        acceleration_index=2,
        n_grid_points=300,
        max_grid=1e7,
        store_progress=False,
    ):
        """
            A synchrotron cooling and acceleration solver and radiation producer with continuous
            injection of power law distributed electrons. The acceleration timescale is taken as a 
            fraction of the cooling timescale. The acceleration index expresses the turbulence distribution
            of the magnetic field: 1 for Bohm, 5/3 for Kolmogorov, 2 for Fermi.


            :param B: the magnetic field strength in Gaussa
            :param index: the injected electron spectral index
            :param gamma_injection: the electron injection energy
            :param gamma_cool: the cooling energy
            :param gamma_max: the maximum injected energy
            :param t_acc_fraction: the acceleration time relative to the cooling time
            :param acceleration_index: the acceleration index
            :param n_grid_points: number of grid points
            :param max_grid: the maximum grid energy
            :param store_progress: to store the progress
            """

        self._B = B

        bulk_gamma = 300.0

        # calculate the number of steps to cool the maximum
        # electron energy to gamma cool

        ratio = gamma_max / gamma_cool

        n_steps = np.round(ratio)

        # assign the properties to the calss

        self._gamma_max = gamma_max
        self._gamma_cool = gamma_cool
        self._gamma_injection = gamma_injection
        self._index = index

        # calculate the cooling time of the highest energy electron
        # and set the acceleration time to the fraction of that

        cooling_time = synchrotron_cooling_time(B, gamma_max)

        t_acc = cooling_time * t_acc_fraction

        # initialize the cooling and acceleration terms

        SynchrotronCoolingAccelerationComponent.__init__(
            self, B, t_acc, acceleration_index
        )

        # create the injection
        ContinuousPowerlawInjection.__init__(
            self, gamma_injection, gamma_max, index, N=1.0
        )

        # make the time step the small enough to account for either the cooling
        # or the acceleration

        delta_t = min(t_acc, cooling_time)

        # initialize the solver

        ChangCooper.__init__(
            self, n_grid_points, max_grid, delta_t, None, store_progress
        )

        # create the synchrotron emission kernel

        emission_kernel = SynchrotronEmission(self._grid, self._B)

        # initialize the photon emission constructor

        PhotonEmitter.__init__(self, n_steps, emission_kernel)



import numpy as np

from pychangcooper.chang_cooper import ChangCooper


class GenericCoolingAccelerationComponent(object):
    def __init__(self, C0=1.0, t_acc=1, cooling_index=-2, acceleration_index=2):

        """
        Generic cooling and acceleration component that must be co-inherited with 
        ChangCooper to form a complete solution. This is to reduce code duplication
        for similar problems
        
        
        :param C0: the cooling constant
        :param t_acc: the acceleration time
        :param cooling_index: the cooling index
        :param acceleration_index: the acceleeration index
        """

        self._C0 = C0
        self._t_acc = t_acc
        self._cooling_index = cooling_index
        self._acceleration_index = acceleration_index

    def _define_terms(self):
        self._dispersion_term = (
            0.5 * np.power(self._half_grid, self._acceleration_index) / self._t_acc
        )

        self._heating_term = (
            self._C0 * np.power(self._half_grid, self._cooling_index)
            - 2 * self._dispersion_term / self._half_grid
        )

    def _t_cool(self, gamma):
        """
        calculate the cooling time of a particle with energy gamma
        :param gamma: the energy of the particle
        :return: 
        """
        return 1.0 / (self._C0 * gamma)

    def _get_min_timescale(self):

        RuntimeWarning("This must be subclassed")

        return self._t_acc

    @property
    def equilbrium_energy(self):
        """
        The equilibrium energy of the electrons
        """

        return 1.0 / (self._C0 * self._t_acc)


class CoolingAcceleration(GenericCoolingAccelerationComponent, ChangCooper):
    def __init__(
        self,
        n_grid_points=300,
        C0=1.0,
        t_acc=1,
        cooling_index=-2,
        acceleration_index=2,
        max_grid=1e7,
        store_progress=False,
        initial_distribution=None,
    ):
        """
        Cooling and acceleration of an unspecified form. 
        
        
        :param n_grid_points: number of grid points 
        :param C0: the cooling constant
        :param t_acc: the acceleration time
        :param cooling_index: the index of the cooling term
        :param acceleration_index: the index of the acceleration term
        :param max_grid: the maximum grid energy
        :param store_progress: to store the progress of the evolution
        :param initial_distribution: the initial distribution of the electrons
        """

        # first call the cool-accel component constructor to setup the terms and special
        # properties

        GenericCoolingAccelerationComponent.__init__(
            self, C0, t_acc, cooling_index, acceleration_index
        )

        # figure out the delta t
        delta_t = self._get_min_timescale()

        ChangCooper.__init__(
            self, n_grid_points, max_grid, delta_t, initial_distribution, store_progress
        )



import numpy as np


class EmissionKernel(object):
    def __init__(self):

        self._photon_energies = None
        self._n_photon_energies = None

    def set_photon_energies(self, photon_energies):

        self._photon_energies = photon_energies
        self._n_photon_energies = len(self._photon_energies)

    @property
    def photon_energies(self):

        return self._photon_energies

    @property
    def n_photon_energies(self):

        return self._n_photon_energies

    def compute_spectrum(self, electron_distribution):

        RuntimeError("Must be implemented in subclass")



import numpy as np
import matplotlib.pyplot as plt

from pychangcooper.io.fill_plot import fill_plot_static
from pychangcooper.tridiagonal_solver import TridiagonalSolver


def log_grid_generator(n_steps, grid_max):
    step = np.exp((1.0 / n_steps) * np.log(grid_max))
    step_plus_one = step + 1.0

    # initialize all the grid points
    grid = np.zeros(n_steps)

    half_grid = np.zeros(n_steps - 1)

    # now build the grid and the half grid points
    # we also make squared terms just incase
    for i in range(n_steps):

        grid[i] = np.power(step, i)

        if i < n_steps - 1:
            half_grid[i] = 0.5 * grid[i] * step_plus_one

    return grid, half_grid, step


class ChangCooper(object):
    def __init__(
        self,
        n_grid_points=300,
        max_grid=1e5,
        delta_t=1.0,
        initial_distribution=None,
        store_progress=False,
    ):
        """
        Generic Chang and Cooper base class. Currently, the dispersion and heating terms 
        are assumed to be time-independent

        :param n_grid_points: number of grid points on the x-axis
        :param max_grid: the maximum energy of the grid
        :param delta_t: the time step in the equation
        :param initial_distribution: an array of an initial electron distribution
        :param store_progress: store the history of the runs
        """

        self._n_grid_points = n_grid_points
        self._dispersion_term = np.zeros(n_grid_points)
        self._heating_term = np.zeros(n_grid_points)

        self._max_grid = max_grid
        self._delta_t = delta_t
        self._iterations = 0
        self._current_time = 0.0
        self._store_progress = store_progress
        self._saved_grids = []

        # first build the grid which is independent of the scheme
        self._build_grid()

        if initial_distribution is None:

            # initalize the grid of electrons
            self._n_current = np.zeros(self._n_grid_points)
            self._initial_distribution = np.zeros(n_grid_points)

        else:

            assert len(initial_distribution) == self._n_grid_points

            self._n_current = np.array(initial_distribution)
            self._initial_distribution = initial_distribution

        # define the heating and dispersion terms
        # must be implemented in the subclasses
        self._define_terms()

        # compute the source/escape function if there is any
        self._compute_source_function_and_escape()

        # compute the delta_js which control the upwind and downwind scheme
        self._compute_delta_j()

        # now compute the tridiagonal terms
        self._setup_vectors()

    def _build_grid(self):
        """
        setup the grid for the calculations and initialize the 
        solution
        """

        # logarithmic grid

        self._grid, self._half_grid, self._step = log_grid_generator(
            self._n_grid_points, self._max_grid
        )
        self._grid2 = self._grid ** 2
        self._half_grid2 = self._half_grid ** 2

        # define the delta of the grid

        # we need to add extra end points to the grid
        # so that we can compute the delta at the boundaries

        first_delta_grid = self._grid[0] * (1 - 1.0 / self._step)
        last_delta_grid = self._grid[-1] * (self._step - 1)

        delta_grid = np.append([first_delta_grid], np.diff(self._grid))
        delta_grid = np.append(delta_grid, [last_delta_grid])

        assert len(delta_grid) == self._n_grid_points + 1

        # delta grid bar is  the average of the grid and applies to
        # to the forward and backwards difference terms
        # When the grid is not uniform, the second derivatives need
        # to be taken in the correct place

        delta_grid_bar = np.mean(np.vstack((delta_grid[:-1], delta_grid[1:])), axis=0)

        assert len(delta_grid_bar) == self._n_grid_points

        self._delta_grid = delta_grid
        self._delta_grid_bar = delta_grid_bar

    def _compute_delta_j(self):
        """
        delta_j controls where the differences are computed. If there are no dispersion
        terms, then delta_j is zero
        """

        # set to zero. note delta_j[n] = 0 by default

        self._delta_j = np.zeros(self._n_grid_points - 1)

        for j in range(self._n_grid_points - 1):

            # if the dispersion term is 0 => delta_j = 0
            if self._dispersion_term[j] != 0:

                w = (
                    self._delta_grid[1:-1][j] * self._heating_term[j]
                ) / self._dispersion_term[j]

                # w asymptotically approaches 1/2, but we need to set it manually
                if w == 0:

                    self._delta_j[j] = 0.5

                # otherwise, we use appropriate bounds
                else:

                    self._delta_j[j] = (1.0 / w) - 1.0 / (np.exp(w) - 1.0)

        # precomoute 1- delta_j
        self._one_minus_delta_j = 1 - self._delta_j

    def _setup_vectors(self):
        """
        from the specified terms in the subclasses, setup the tridiagonal terms
        
        """

        # initialize everything to zero

        a = np.zeros(self._n_grid_points)
        b = np.zeros(self._n_grid_points)
        c = np.zeros(self._n_grid_points)

        # walk backwards in j starting from the second to last index
        # then set the end points
        for k in range(self._n_grid_points - 2, 0, -1):
            # pre compute one over the delta of the grid
            # this is the 1/delta_grid in front of the F_j +/- 1/2.

            one_over_delta_grid_forward = 1.0 / self._delta_grid[k + 1]
            one_over_delta_grid_backward = 1.0 / self._delta_grid[k]

            # this is the delta grid in front of the full equation

            one_over_delta_grid_bar = 1.0 / self._delta_grid_bar[k]

            # The B_j +/- 1/2 from CC
            B_forward = self._heating_term[k]
            B_backward = self._heating_term[k - 1]

            # The C_j +/- 1/2 from CC
            C_forward = self._dispersion_term[k]
            C_backward = self._dispersion_term[k - 1]

            # in order to keep math errors at a minimum, the tridiagonal terms
            # are computed in separate functions so that boundary conditions are
            # set consistently.

            # First we solve (N - N) = F
            # then we will move the terms to form a tridiagonal equation

            # n_j-1 term
            a[k] = _compute_n_j_minus_one_term(
                one_over_delta_grid=one_over_delta_grid_bar,
                one_over_delta_grid_bar_backward=one_over_delta_grid_backward,
                C_backward=C_backward,
                B_backward=B_backward,
                delta_j_minus_one=self._delta_j[k - 1],
            )

            # n_j term
            b[k] = _compute_n_j(
                one_over_delta_grid=one_over_delta_grid_bar,
                one_over_delta_grid_bar_backward=one_over_delta_grid_backward,
                one_over_delta_grid_bar_forward=one_over_delta_grid_forward,
                C_backward=C_backward,
                C_forward=C_forward,
                B_backward=B_backward,
                B_forward=B_forward,
                one_minus_delta_j_minus_one=self._one_minus_delta_j[k - 1],
                delta_j=self._delta_j[k],
            )

            # n_j+1 term
            c[k] = _compute_n_j_plus_one(
                one_over_delta_grid=one_over_delta_grid_bar,
                one_over_delta_grid_bar_forward=one_over_delta_grid_forward,
                C_forward=C_forward,
                B_forward=B_forward,
                one_minus_delta_j=self._one_minus_delta_j[k],
            )

        # now set the end points

        ################
        # right boundary
        # j+1/2 = 0

        one_over_delta_grid_forward = 0.0
        one_over_delta_grid_backward = 1.0 / self._delta_grid[-1]

        one_over_delta_grid_bar = 1.0 / self._delta_grid_bar[-1]

        # n_j-1 term
        a[-1] = _compute_n_j_minus_one_term(
            one_over_delta_grid=one_over_delta_grid_bar,
            one_over_delta_grid_bar_backward=one_over_delta_grid_backward,
            C_backward=self._dispersion_term[-1],
            B_backward=self._heating_term[-1],
            delta_j_minus_one=self._delta_j[-1],
        )

        # n_j term
        b[-1] = _compute_n_j(
            one_over_delta_grid=one_over_delta_grid_bar,
            one_over_delta_grid_bar_backward=one_over_delta_grid_backward,
            one_over_delta_grid_bar_forward=one_over_delta_grid_forward,
            C_backward=self._dispersion_term[-1],
            C_forward=0,
            B_backward=self._heating_term[-1],
            B_forward=0,
            one_minus_delta_j_minus_one=self._one_minus_delta_j[-1],
            delta_j=0,
        )

        # n_j+1 term
        c[-1] = 0

        ###############
        # left boundary
        # j-1/2 = 0

        one_over_delta_grid = 1.0 / (
            self._half_grid[0] - self._grid[0] / np.sqrt(self._step)
        )

        one_over_delta_grid_bar_forward = 1.0 / self._delta_grid_bar[0]
        one_over_delta_grid_bar_backward = 0.0

        one_over_delta_grid_forward = 1.0 / self._delta_grid[0]
        one_over_delta_grid_backward = 0

        one_over_delta_grid_bar = 1.0 / self._delta_grid_bar[0]

        # n_j-1 term
        a[0] = 0.0

        # n_j term
        b[0] = _compute_n_j(
            one_over_delta_grid=one_over_delta_grid_bar,
            one_over_delta_grid_bar_backward=one_over_delta_grid_backward,
            one_over_delta_grid_bar_forward=one_over_delta_grid_forward,
            C_backward=0,
            C_forward=self._dispersion_term[0],
            B_backward=0,
            B_forward=self._heating_term[0],
            one_minus_delta_j_minus_one=0,
            delta_j=self._delta_j[0],
        )
        # n_j+1 term
        c[0] = _compute_n_j_plus_one(
            one_over_delta_grid=one_over_delta_grid_bar,
            one_over_delta_grid_bar_forward=one_over_delta_grid_forward,
            C_forward=self._dispersion_term[0],
            B_forward=self._heating_term[0],
            one_minus_delta_j=self._one_minus_delta_j[0],
        )

        # carry terms to the other side to form a tridiagonal equation
        # the escape term is added on but is zero unless created in
        # a child class

        a *= -self._delta_t
        b = (1 - b * self._delta_t) + self._escape_grid * self._delta_t
        c *= -self._delta_t

        # now make a tridiagonal_solver for these terms

        self._tridiagonal_solver = TridiagonalSolver(a, b, c)

    def _compute_source_function_and_escape(self):
        """
        compute the grid of the source term. This will just be zero if there is nothing 
        added and all will be zero
        """

        self._source_grid = self._source_function(self._grid)
        self._escape_grid = self._escape_function(self._grid)

    def _source_function(self, energy):

        return 0.0

    def _escape_function(self, energy):

        return 0

    def _define_terms(self):

        raise RuntimeError("Must be implemented in subclass")

    def solve_time_step(self):
        """
        Solve for the next time step. 
        """

        # if we are storing the solutions, then append them
        # to the history

        if self._store_progress:
            self._saved_grids.append(self._n_current)

        # set up the right side of the tridiagonal equation.
        # This is the current distribution plus the source
        # unless it is zero

        d = self._n_current + self._source_grid * self._delta_t

        # set the new solution to the current one

        self._n_current = self._tridiagonal_solver.solve(d)

        # clean any numerical diffusion if needed
        # this must be customized

        self._clean()

        # bump up the iteration number and the time

        self._iteratate()

    def _clean(self):

        pass

    def _iteratate(self):
        """
        increase the run iterator and the current time
        """

        # bump up the iteration number

        self._iterations += 1

        # increase the time

        self._current_time += self._delta_t

    @property
    def current_time(self):
        """
        The current time: delta_t * n_iterations
        """

        return self._current_time

    @property
    def n_iterations(self):
        """
        The number of iterations solved for
        """

        return self._iterations

    @property
    def delta_j(self):
        """
        the delta_js 
        """

        return self._delta_j

    @property
    def grid(self):
        """
        The energy grid
        """

        return self._grid

    @property
    def half_grid(self):
        """
        The half energy grid
        """

        return self._half_grid

    @property
    def n(self):
        """
        The current solution
        """

        return self._n_current

    @property
    def history(self):
        """
        The history of the solution
        """

        return np.array(self._saved_grids)

    def reset(self):
        """
        reset the solver to the initial electron distribution
        
        :return: 
        """

        self._n_current = self._initial_distribution
        self._iterations = 0
        self._current_time = 0.0

    def plot_current_distribution(self, ax=None, **kwargs):

        if ax is None:

            fig, ax = plt.subplots()

            ax.set_xlabel(r"$\gamma$")
            ax.set_ylabel(r"$N(\gamma, t)$")

            ax.set_xscale("log")
            ax.set_yscale("log")

        else:

            fig = ax.get_figure()

        ax.plot(self._grid, self._n_current, **kwargs)

        return fig

    def plot_initial_distribution(self, ax=None, **kwargs):
        """
        plot the initial distribution of the electrons
        
        :param ax: ax to plot to
        :param kwargs: mpl kwargs
        :return: fig
        """

        if ax is None:

            fig, ax = plt.subplots()

            ax.set_xlabel(r"$\gamma$")
            ax.set_ylabel(r"$N(\gamma, t)$")

            ax.set_xscale("log")
            ax.set_yscale("log")

        else:

            fig = ax.get_figure()

        ax.plot(self._grid, self._initial_distribution, **kwargs)

        return fig

    def plot_evolution(
        self,
        cmap="magma",
        skip=1,
        show_legend=False,
        alpha=0.9,
        show_final=False,
        show_initial=False,
        ax=None,
    ):
        """
        plot th evolution of the electrons
        
        
        :param cmap: cmap to use
        :param skip: number of elements to skip
        :param show_legend: show a legend
        :param alpha: the transparency 
        :param show_final: label the final solution
        :param show_initial: show the initial distribution
        :param ax: the ax to plot to
        :return: 
        """

        solutions = self.history[::skip]

        fig = fill_plot_static(self._grid, solutions, cmap, alpha, ax)

        if ax is None:
            ax = fig.get_axes()[0]

        if show_final:
            _ = self.plot_current_distribution(
                ax,
                color="k",
                ls="--",
                zorder=len(solutions) + 1,
                alpha=1,
                label="final distribution",
            )

        if show_initial:
            _ = self.plot_initial_distribution(
                ax,
                color="k",
                ls=":",
                zorder=len(solutions) + 1,
                alpha=1,
                label="initial distribution",
            )

        ax.set_xscale("log")
        ax.set_yscale("log")

        ax.set_xlabel(r"$\gamma$")
        ax.set_ylabel(r"$N(\gamma$,t)")

        if show_legend:
            ax.legend()

        return fig


def _compute_n_j_plus_one(
    one_over_delta_grid,
    one_over_delta_grid_bar_forward,
    C_forward,
    B_forward,
    one_minus_delta_j,
):
    """
    equation for the CC n_j +1 term

    :param one_over_delta_grid: the total change in energy
    :param one_over_delta_grid_bar_backward: the backward change in energy for the second derivative
    :param C_forward: the forward dispersion term
    :param B_forward: the forward heating term
    :param one_minus_delta_j: 1 - delta_j
    """

    return one_over_delta_grid * (
        one_minus_delta_j * B_forward + one_over_delta_grid_bar_forward * C_forward
    )


def _compute_n_j(
    one_over_delta_grid,
    one_over_delta_grid_bar_backward,
    one_over_delta_grid_bar_forward,
    C_backward,
    C_forward,
    B_backward,
    B_forward,
    one_minus_delta_j_minus_one,
    delta_j,
):
    """
    equation for the CC n_j term

    :param one_over_delta_grid: the total change in energy
    :param one_over_delta_grid_bar_backward: the backward change in energy for the second derivative
    :param one_over_delta_grid_bar_forward: the forward change in energy for the second derivative
    :param C_forward: the forward dispersion term
    :param C_backward: the backward dispersion term
    :param B_forward: the forward heating term
    :param B_backward: the backward heating term
    :param one_minus_delta_j_minus_one: 1 - delta_j-1
    """

    return -one_over_delta_grid * (
        (
            one_over_delta_grid_bar_forward * C_forward
            + one_over_delta_grid_bar_backward * C_backward
        )
        + one_minus_delta_j_minus_one * B_backward
        - delta_j * B_forward
    )


def _compute_n_j_minus_one_term(
    one_over_delta_grid,
    one_over_delta_grid_bar_backward,
    C_backward,
    B_backward,
    delta_j_minus_one,
):
    """
    equation for the CC n_j-1 term

    :param one_over_delta_grid: the total change in energy
    :param one_over_delta_grid_bar_forward: the forward change in energy for the second derivative
    :param C_backward: the backward dispersion term
    :param B_backward: the backward heating term
    :param one_minus_delta_j: 1 - delta_j
    """

    return one_over_delta_grid * (
        one_over_delta_grid_bar_backward * C_backward - delta_j_minus_one * B_backward
    )



from pychangcooper.chang_cooper import ChangCooper
from pychangcooper.scenarios.synchrotron_cooling import (
    SynchrotronCooling_ContinuousPLInjection,
    SynchrotronCooling_ImpulsivePLInjection,
)
from pychangcooper.scenarios.synchrotron_cooling_acceleration import (
    SynchCoolAccel_ImpulsivePLInjection,
    SynchCoolAccel_ContinuousPLInjection,
)

from pychangcooper.scenarios.generic_cooling_acceleration import CoolingAcceleration

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



import numpy as np
from numba import jit, njit


@njit(fastmath=True)
def jit_forward_sweep(n, cprime, dprime, a, b, c, d):
    for i in range(1, n):
        b_minus_ac = b[i] - a[i] * cprime[i - 1]

        cprime[i] = c[i] / b_minus_ac

        dprime[i] = (d[i] - a[i] * dprime[i - 1]) / b_minus_ac


@njit(fastmath=True)
def jit_backward_sweep(n, n_j_plus_1, cprime, dprime):
    for j in range(n - 2, -1, -1):
        n_j_plus_1[j] = dprime[j] - cprime[j] * n_j_plus_1[j + 1]


class TridiagonalSolver(object):
    def __init__(self, a, b, c):
        """
        A tridiagonal solver for the equation:

        a_i x_i-1 + b_i x_i _ c_i x_i+1 = d_i

        :param a: the x_i-1 terms
        :param b: the x_i terms
        :param c: the x_i+1 terms


        """

        # get the number of elements
        self._n_grid_points = len(a)

        # make sure the elements are all the same
        # size
        assert len(a) == len(b)
        assert len(a) == len(c)

        # assign the elements to the class
        self._a = np.array(a)
        self._b = np.array(b)
        self._c = np.array(c)

        # if the a terms are all zero, we do not need the
        # forward sweep as a is already eliminated

        self._a_non_zero = ~np.all(self._a == 0)

        # the 0th term of the c prime will always
        # be the same so pre compute.
        # is a = 0 then all terms will be the same.

        # this is basically the first step of the forward
        # sweep

        self._cprime = self._c / self._b

    def _forward_sweep(self, d):
        """
        This is the forward sweep of the tridiagonal solver


        """

        # we have already set the first terms in c prime
        # if we need to forward sweep, we must set the remaining
        # terms. Otherwise, they are just ratios

        self._cprime = self._c / self._b
        self._dprime = d / self._b

        if self._a_non_zero:
            jit_forward_sweep(
                self._n_grid_points,
                self._cprime,
                self._dprime,
                self._a,
                self._b,
                self._c,
                d,
            )

    def _backwards_substitution(self):
        """
        This is the backwards substitution step of the tridiagonal
        solver. 
        """

        n_j_plus_1 = np.zeros(self._n_grid_points)

        # set the end points
        n_j_plus_1[-1] = self._dprime[-1]

        # backwards step to the beginning

        jit_backward_sweep(self._n_grid_points, n_j_plus_1, self._cprime, self._dprime)

        return n_j_plus_1

    def solve(self, d_j):
        """
        
        """

        self._forward_sweep(d_j)

        d_j_plus_one = self._backwards_substitution()

        return d_j_plus_one
